/*
 * search.c  -- isearch
 *
 * $Id: search.c,v 1.7 2004/06/01 05:03:07 hos Exp $
 *
 */


#include "main.h"
#include "util.h"
#include <errno.h>


#define WM_APP_FWD (WM_APP + 1)
#define WM_APP_BWD (WM_APP + 2)
#define WM_APP_CLOSE (WM_APP + 3)
#define WM_APP_SEL_ACTION (WM_APP + 4)

#define MIGEMO_ENV "MIGEMO_DICT"


typedef LRESULT (*msg_proc_t)(HWND, UINT, WPARAM, LPARAM, wnd_ctx_t *);


static TCHAR *window_class_name = _T("migemo-isearch");
static TCHAR *window_title_name = _T("migemo-isearch");
static int margin = 4;


static HRESULT get_selected_range(wnd_ctx_t *ctx, IDispatch **range)
{
    HRESULT hres;

    hres = ctx->selected_range(ctx, range);
    if(FAILED(hres)) {
        return hres;
    }

    return S_OK;
}

static HRESULT get_selected_text(wnd_ctx_t *ctx, BSTR *word)
{
    IDispatch *range;
    HRESULT hres;

    hres = get_selected_range(ctx, &range);
    if(FAILED(hres)) {
        return hres;
    }

    hres = get_property_strv(range, ctx->range_text, word);
    IDispatch_Release(range);
    if(FAILED(hres)) {
        return hres;
    }

    return S_OK;
}


static unsigned char *mbsdup_from_bstr(BSTR str)
{
    int size;
    unsigned char *mbs;

    size = WideCharToMultiByte(CP_OEMCP, 0,
                               str, wcslen(str), NULL, 0,
                               NULL, NULL);

    mbs = (unsigned char *)malloc(size + 1);
    if(mbs == NULL) {
        return NULL;
    }
    memset(mbs, 0, size + 1);

    WideCharToMultiByte(CP_OEMCP, 0,
                        str, wcslen(str), mbs, size + 1,
                        NULL, NULL);

    return mbs;
}

static BSTR bstrdup_from_mbs(unsigned char *mbs)
{
    int size;
    BSTR str;

    size = MultiByteToWideChar(CP_OEMCP, 0, mbs, strlen(mbs), NULL, 0);

    str = SysAllocStringLen(NULL, size);
    if(str == NULL) {
        return NULL;
    }

    MultiByteToWideChar(CP_OEMCP, 0, mbs, strlen(mbs), str, size + 1);

    return str;
}

static int is_ascii_str(BSTR str)
{
    while(str[0] != 0) {
        if(! iswascii(str[0])) {
            return 0;
        }

        str += 1;
    }

    return 1;
}


static HRESULT find_text(wnd_ctx_t *ctx, IDispatch *range, VARIANT_BOOL *res)
{
    IMatchCollection *matches;
    HRESULT hres;

    if(ctx->mode & MODE_NORMAL) {
        while(1) {
            VARIANT_BOOL find_res;
            IDispatch *org_range;

            hres = get_property_dpv(range, ctx->range_dup, &org_range);
            if(FAILED(hres)) {
                return hres;
            }

            hres = ctx->range_find_proc(range, ctx, ctx->word, &find_res);
            if(FAILED(hres) || find_res == VARIANT_FALSE) {
                IDispatch_Release(org_range);
                *res = VARIANT_FALSE;
                return S_OK;
            }

            hres = ctx->range_select(ctx, range);
            if(SUCCEEDED(hres)) {
                IDispatch_Release(org_range);
                *res = VARIANT_TRUE;
                return S_OK;
            }

            hres = ctx->range_move_point(org_range, range, ctx->dir, ctx->dir);
            if(FAILED(hres)) {
                IDispatch_Release(org_range);
                return hres;
            }

            IDispatch_Release(range);
            range = org_range;
        }
    } else if((ctx->mode & MODE_MIGEMO) && is_ascii_str(ctx->word)) {
        unsigned char *word_mb, *pattern_mb;
        BSTR pattern;

        word_mb = mbsdup_from_bstr(ctx->word);
        if(word_mb == NULL) {
            return E_FAIL;
        }

        pattern_mb = migemo_query_p(ctx->mp, word_mb);
        free(word_mb);
        if(pattern_mb == NULL) {
            return E_FAIL;
        }

        pattern = bstrdup_from_mbs(pattern_mb);
        migemo_release_p(ctx->mp, pattern_mb);
        if(pattern == NULL) {
            return E_FAIL;
        }

        hres = IRegExp_put_Pattern(ctx->regexp, pattern);
        SysFreeString(pattern);
        if(FAILED(hres)) {
            return hres;
        }
    } else {
        hres = IRegExp_put_Pattern(ctx->regexp, ctx->word);
        if(FAILED(hres)) {
            return hres;
        }
    }

    {
        BSTR text;
        IDispatch *matches_dp;

        hres = get_property_strv(range, ctx->range_text, &text);
        if(FAILED(hres)) {
            return hres;
        }

        hres = IRegExp_Execute(ctx->regexp, text, &matches_dp);
        SysFreeString(text);
        if(FAILED(hres)) {
            *res = VARIANT_FALSE;
            return S_OK;
        }

        hres = IDispatch_QueryInterface(matches_dp, &IID_IMatchCollection,
                                        (void **)(void *)&matches);
        IDispatch_Release(matches_dp);
        if(FAILED(hres)) {
            return hres;
        }
    }

    {
        long cnt, i, idx;
        IDispatch *match_dp;
        IMatch *match;
        BSTR match_str;
        VARIANT_BOOL find_res;

        hres = IMatchCollection_get_Count(matches, &cnt);
        if(FAILED(hres)) {
            IMatchCollection_Release(matches);
            return hres;
        }

        for(i = 0; i < cnt; i++) {
            idx = (ctx->dir > 0 ? i : cnt - i - 1);

            hres = IMatchCollection_get_Item(matches, idx, &match_dp);
            if(FAILED(hres)) {
                IMatchCollection_Release(matches);
                return hres;
            }

            hres = IMatchCollection_QueryInterface(match_dp, &IID_IMatch,
                                                   (void **)(void *)&match);
            IDispatch_Release(match_dp);
            if(FAILED(hres)) {
                IMatchCollection_Release(matches);
                return hres;
            }

            hres = IMatch_get_Value(match, &match_str);
            IMatch_Release(match);
            if(FAILED(hres)) {
                IMatchCollection_Release(matches);
                return hres;
            }

            {
                IDispatch *org_range;

                hres = get_property_dpv(range, ctx->range_dup, &org_range);
                if(FAILED(hres)) {
                    IMatchCollection_Release(matches);
                    return hres;
                }

                hres = ctx->range_find_proc(range, ctx, match_str, &find_res);
                SysFreeString(match_str);
                if(FAILED(hres) || find_res == VARIANT_FALSE) {
                    IDispatch_Release(org_range);
                } else {
                    hres = ctx->range_select(ctx, range);
                    if(SUCCEEDED(hres)) {
                        IDispatch_Release(org_range);
                        *res = VARIANT_TRUE;
                        break;
                    }

                    hres = ctx->range_move_point(org_range, range,
                                                 ctx->dir, ctx->dir);
                    if(FAILED(hres)) {
                        IDispatch_Release(org_range);
                        IMatchCollection_Release(matches);
                        return hres;
                    }

                    IDispatch_Release(range);
                    range = org_range;
                }
            }
        }
        if(i == cnt) {
            *res = VARIANT_FALSE;
        }
    }

    IMatchCollection_Release(matches);

    return S_OK;
}


static void set_status_text(wnd_ctx_t *ctx)
{
    TCHAR buf1[256], buf2[256];

    wsprintf(buf1, _T("%s%s%sI-search%s:"),
             (ctx->failing ? "failing " : ""),
             (ctx->wrapped ? "wrapped " : ""),
             ((ctx->mode & MODE_REGEXP) ? "regexp " : ""),
             (ctx->dir > 0 ? "" : " backward"));
    buf1[0] = _totupper(buf1[0]);

    wsprintf(buf2, _T("%s%s"),
             ((ctx->mode & MODE_MIGEMO) ? "[MIGEMO] " : ""),
             buf1);

    SetWindowText(ctx->stat, buf2);
}

static void search_word(wnd_ctx_t *ctx, int offset)
{
    IDispatch *search_range = NULL;
    HRESULT hres;

    if(wcslen(ctx->word) == 0) {
        return;
    }

    hres = get_property_dpv(ctx->doc, ctx->doc_all_range, &search_range);
    if(FAILED(hres)) {
        error_message("get doc all range failed", hres);
        return;
    }

    if(ctx->failing) {
        ctx->failing = FALSE;
    } else {
        IDispatch *sel_range;

        hres = get_selected_range(ctx, &sel_range);
        if(SUCCEEDED(hres)) {
            hres = ctx->range_move_point(search_range, sel_range,
                                         ctx->dir, offset);
            IDispatch_Release(sel_range);
            if(FAILED(hres)) {
                error_message("ctx->range_move_point failed", hres);
                goto func_end;;
            }
        }
    }

    {
        VARIANT_BOOL res;

        hres = find_text(ctx, search_range, &res);
        if(FAILED(hres)) {
            error_message("find_text failed", hres);
            goto func_end;
        }

        if(res == VARIANT_FALSE) {
            ctx->failing = 1;
        }
    }

    set_status_text(ctx);

  func_end:

    if(search_range != NULL) IDispatch_Release(search_range);

}


static LRESULT create_proc(HWND hwnd, UINT msg,
                           WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    BSTR word;
    int w, h;

    {
        RECT rect;

        GetClientRect(hwnd, &rect);
        w = rect.right - margin * 2;
        h = rect.bottom - margin * 2;
    }

    ctx->stat = CreateWindow(_T("Static"), _T(""),
                             WS_VISIBLE | WS_CHILD,
                             margin, margin, (int)(w * 0.4), h,
                             hwnd, NULL, NULL, NULL);
    if(ctx->stat == NULL) {
        return -1;
    }

    ctx->edit = CreateWindowW(L"Edit", L"",
                              WS_VISIBLE | WS_CHILD,
                              margin + (int)(w * 0.4), margin,
                              (int)(w * 0.6), h,
                              hwnd, NULL, NULL, NULL);
    if(ctx->edit == NULL) {
        return -1;
    }

    if(FAILED(get_selected_text(ctx, &word)) || word == NULL) {
        word = SysAllocString(L"");
        if(word == NULL) {
            return -1;
        }
    }
    ctx->word = word;

    SendMessage(ctx->stat, WM_SETFONT,
                (WPARAM)GetStockObject(DEFAULT_GUI_FONT), 0);
    set_status_text(ctx);

    SendMessageW(ctx->edit, WM_SETFONT,
                 (WPARAM)GetStockObject(DEFAULT_GUI_FONT), 0);
    SetWindowTextW(ctx->edit, ctx->word);
    SetFocus(ctx->edit);
    if(wcslen(ctx->word) != 0) {
        SendMessageW(ctx->edit, EM_SETSEL, 0, -1);
    }

    ctx->main = hwnd;

    return 0;
}

static LRESULT destroy_proc(HWND hwnd, UINT msg,
                            WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    SysFreeString(ctx->word);

    return 0;
}

static LRESULT close_proc(HWND hwnd, UINT msg,
                          WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    DestroyWindow(hwnd);
    ctx->main = NULL;

    return 0;
}

static LRESULT edit_change(HWND hwnd, UINT msg,
                           WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    BSTR word;
    int len;

    if(! (ctx->mode & MODE_INCREMENTAL)) {
        return 0;
    }

    if(ctx->main == NULL) {
        return 0;
    }

    len = GetWindowTextLengthW(ctx->edit);
    word = SysAllocStringLen(NULL, len);
    if(word == NULL) {
        return 0;
    }

    GetWindowTextW(ctx->edit, word, len + 1);

    SysFreeString(ctx->word);
    ctx->word = word;

    ctx->failing = FALSE;
    if(ctx->dir > 0) {
        search_word(ctx, 0);
    } else {
        search_word(ctx, +1);
    }

    return 0;
}

static LRESULT edit_killfocus(HWND hwnd, UINT msg,
                              WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    PostMessage(ctx->main, WM_APP_CLOSE, 0, 0);

    return 0;
}

static LRESULT command_proc(HWND hwnd, UINT msg,
                            WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    WORD code;
    msg_proc_t proc;

    static struct uint_ptr_pair cmd_proc[] = {
        {EN_CHANGE, edit_change},
        {EN_KILLFOCUS, edit_killfocus},

        {0, NULL}
    };

    code = HIWORD(wparam);

    proc = (msg_proc_t)assq_pair(cmd_proc, code, NULL);
    if(proc == NULL) {
        return 0;
    }

    return proc(hwnd, msg, wparam, lparam, ctx);
}

static LRESULT forward_proc(HWND hwnd, UINT msg,
                            WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    if(ctx->dir > 0 && ctx->failing) {
        ctx->wrapped = TRUE;
    }
    ctx->dir = +1;
    search_word(ctx, +1);

    return 0;
}

static LRESULT backward_proc(HWND hwnd, UINT msg,
                             WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    if(ctx->dir < 0 && ctx->failing) {
        ctx->wrapped = TRUE;
    }
    ctx->dir = -1;
    search_word(ctx, -1);

    return 0;
}

static LRESULT sel_action(HWND hwnd, UINT msg,
                          WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    IDispatch *sel_range;
    HRESULT hres;

    hres = get_selected_range(ctx, &sel_range);
    if(FAILED(hres)) {
        error_message("get_selected_range failed", hres);
        return 0;
    }

    hres = ctx->range_action(ctx, sel_range);
    IDispatch_Release(sel_range);
    if(FAILED(hres)) {
        error_message("ctx->range_action failed", hres);
        return 0;
    }

    return 0;
}

static LRESULT default_proc(HWND hwnd, UINT msg,
                            WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    return DefWindowProc(hwnd, msg, wparam, lparam);
}

static LRESULT CALLBACK window_proc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam)
{
    msg_proc_t proc;
    wnd_ctx_t *ctx;

    static struct uint_ptr_pair msg_proc[] = {
        {WM_CREATE, create_proc},
        {WM_DESTROY, destroy_proc},
        {WM_APP_CLOSE, close_proc},
        {WM_COMMAND, command_proc},
        {WM_APP_FWD, forward_proc},
        {WM_APP_BWD, backward_proc},
        {WM_APP_SEL_ACTION, sel_action},

        {0, NULL}
    };

    if(msg == WM_CREATE) {
        CREATESTRUCT *cs;

        cs = (CREATESTRUCT *)lparam;
        ctx = (wnd_ctx_t *)cs->lpCreateParams;
        SetWindowLongPtr(hwnd, 0, (LONG_PTR)ctx);
    } else {
        ctx = (wnd_ctx_t *)GetWindowLongPtr(hwnd, 0);
    }

    proc = (msg_proc_t)assq_pair(msg_proc, msg, default_proc);

    return proc(hwnd, msg, wparam, lparam, ctx);
}


static HWND create_window(wnd_ctx_t *param, HWND parent)
{
    HWND hwnd;
    int ret;

    {
        WNDCLASSEX wc;

        memset(&wc, 0, sizeof(wc));
        wc.cbSize = sizeof(wc);
        wc.style = 0;
        wc.lpfnWndProc = window_proc;
        wc.cbClsExtra = 0;
        wc.cbWndExtra = sizeof(wnd_ctx_t *);
        wc.hInstance = instance;
        wc.hIcon = NULL;
        wc.hCursor = NULL;
        wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
        wc.lpszMenuName = NULL;
        wc.lpszClassName = window_class_name;
        wc.hIconSm = NULL;

        ret = RegisterClassEx(&wc);
        if(ret == 0) {
            return NULL;
        }
    }

    {
        RECT rect;
        int x, y, w, h;
        LOGFONT logfont;

        GetObject(GetStockObject(DEFAULT_GUI_FONT), sizeof(logfont), &logfont);
        h = logfont.lfHeight;
        if(h < 0) {
            HDC dc;

            dc = GetDC(NULL);
            h = -MulDiv(h, GetDeviceCaps(dc, LOGPIXELSY), 72);
            ReleaseDC(NULL, dc);
        }

        h = h + margin * 2;

        GetWindowRect(parent, &rect);
        w = rect.right - rect.left;
        x = rect.left;
        y = rect.bottom - h;

        hwnd = CreateWindow(window_class_name, window_title_name,
                            WS_POPUP,
                            x, y, w, h,
                            parent, NULL, instance, param);
    }

    return hwnd;
}

static void message_loop(wnd_ctx_t *ctx)
{
    MSG msg;
    int ret;

    while(ctx->main != NULL) {
        ret = GetMessage(&msg, NULL, 0, 0);
        if(ret < 0) {
            break;
        }

        if(ret == 0) {
            break;
        }

        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
}


static void isearch_doc(wnd_ctx_t *ctx)
{
    HRESULT hres;

    if(ctx->mode & (MODE_REGEXP | MODE_MIGEMO)) {
        /* IRegExp */
        hres = CoCreateInstance(&CLSID_RegExp, NULL, CLSCTX_INPROC_SERVER,
                                &IID_IRegExp, (void **)&ctx->regexp);
        if(FAILED(hres)) {
            error_message("CoCreateInstance failed", hres);
            goto func_end;
        }

        hres = IRegExp_put_Global(ctx->regexp, VARIANT_TRUE);
        if(FAILED(hres)) {
            error_message("IRegExp_put_Global failed", hres);
            goto func_end;
        }

        hres = IRegExp_put_IgnoreCase(ctx->regexp,
                                      ((ctx->mode & MODE_IGNORECASE) ?
                                       VARIANT_TRUE : VARIANT_FALSE));
        if(FAILED(hres)) {
            error_message("IRegExp_put_IgnoreCase failed", hres);
            goto func_end;
        }
    }

    /* main window */
    {
        HWND hwnd;

        hwnd = create_window(ctx, ctx->parent);
        if(hwnd == NULL) {
            error_message("create_window failed", GetLastError());
            goto func_end;
        }

        ShowWindow(hwnd, SW_SHOW);
    }

    /* migemo */
    if(ctx->mode & MODE_MIGEMO) {
        if(migemo_open_p == NULL) {
            error_message("migemo library is not found", 0);
            goto func_end;
        }

        ctx->mp = migemo_open_p(ctx->migemo_dict);
        if(ctx->mp == NULL) {
            error_message("migemo_open failed", errno);
            goto func_end;
        }

        if(! migemo_is_enable_p(ctx->mp)) {
            error_message("failed to load migemo dictionary", 0);
            goto func_end;
        }
    }

    /* main loop */
    message_loop(ctx);

  func_end:
    if(ctx->main != NULL) DestroyWindow(ctx->main);
    if(ctx->regexp != NULL) IRegExp_Release(ctx->regexp);
    if(ctx->mp != NULL) migemo_close_p(ctx->mp);
    UnregisterClass(window_class_name, instance);
}


void isearch(wnd_ctx_t *ctx)
{
    HRESULT hres;
    int migemo_dict_allocated = 0;

    hres = CoInitialize(NULL);
    if(FAILED(hres)) {
        error_message("CoInitialize failed", hres);
        return;
    }

    if(strcmp(ctx->migemo_dict, "") == 0) {
        DWORD len;

        len = GetEnvironmentVariableA(MIGEMO_ENV, NULL, 0);
        if(len > 0) {
            char *buf;

            buf = malloc(len);
            if(buf == NULL) {
                error_message("malloc failed", errno);
                goto func_end;
            }

            GetEnvironmentVariableA(MIGEMO_ENV, buf, len);
            ctx->migemo_dict = buf;
            migemo_dict_allocated = 1;
        } else {
            ctx->migemo_dict = "./dict/migemo-dict";
        }
    }

    ctx->parent = GetForegroundWindow();
    if(ctx->parent == NULL) {
        error_message("GetForegroundWindow failed", GetLastError());
        goto func_end;
    }

    hres = ctx->get_doc(&ctx->doc);
    if(FAILED(hres)) {
        error_message("get_doc failed", hres);
        goto func_end;
    }

    isearch_doc(ctx);

  func_end:
    if(migemo_dict_allocated) free(ctx->migemo_dict);
    CoUninitialize();
}
