/*
 * word.c  -- Word depend
 *
 * $Id: word.c,v 1.4 2004/05/24 11:16:26 hos Exp $
 *
 */


#include "main.h"


static HRESULT word_selected_range(wnd_ctx_t *ctx, IDispatch **sel_range)
{
    LPOLESTR doc_selection_range[] =
        {L"Application", L"Selection", L"Range", NULL};

    return get_property_dpv(ctx->doc, doc_selection_range, sel_range);
}

static HRESULT word_range_find(IDispatch *range, wnd_ctx_t *ctx, BSTR text,
                               VARIANT_BOOL *res)
{
    IDispatch *find;
    VARIANT var;
    HRESULT hres;

    hres = get_property_dp(range, L"Find", &find);
    if(FAILED(hres)) {
        return hres;
    }

    hres = call_method_s(find, L"ClearFormatting", NULL, 0, NULL);
    if(FAILED(hres)) {
        goto func_end;
    }

    var.vt = VT_BSTR;
    var.bstrVal = text;
    hres = put_property_s(find, L"Text", &var);
    if(FAILED(hres)) {
        goto func_end;
    }

    hres = call_method_s(find, L"Execute", NULL, 0, &var);
    if(FAILED(hres)) {
        goto func_end;
    }
    if(var.vt != VT_BOOL) {
        hres = E_FAIL;
        goto func_end;
    }

    *res = var.boolVal;

  func_end:
    IDispatch_Release(find);

    return S_OK;
}

static HRESULT word_range_move_point(IDispatch *search_range,
                                     IDispatch *sel_range,
                                     int dir, int offset)
{
    VARIANT args[2];
    long pos;
    BSTR prop, move_method;
    HRESULT hres;

    prop = (dir > 0 ? L"Start" : L"End");
    move_method = (dir > 0 ? L"MoveStart" : L"MoveEnd");

    hres = get_property_long(sel_range, prop, &pos);
    IDispatch_Release(sel_range);
    if(FAILED(hres)) {
        return hres;
    }

    hres = put_property_long(search_range, prop, pos);
    if(FAILED(hres)) {
        return hres;
    }

    args[0].vt = VT_I4;
    args[0].lVal = offset;
    args[1].vt = VT_I4;
    args[1].lVal = 2;  /* WdUnits.wdCharacter */
    hres = call_method_s(search_range, move_method, args, 2, NULL);
    if(FAILED(hres)) {
        return hres;
    }

    return S_OK;
}

static HRESULT word_range_select(wnd_ctx_t *ctx, IDispatch *range)
{
    HRESULT hres;

    hres = call_method_s(range, L"Select", NULL, 0, NULL);
    if(FAILED(hres)) {
        return hres;
    }

    return S_OK;
}


static HRESULT get_word_doc(IDispatch **doc)
{
    return get_office_doc(doc,
                          L"Word.Application",
                          L"ActiveDocument");
}


static LPOLESTR range_text[] = {L"Text", NULL};
static LPOLESTR range_dup[] = {L"Duplicate", NULL};
static LPOLESTR doc_all_range[] = {L"Content", NULL};

static HRESULT word_init_context(wnd_ctx_t *ctx)
{
    ctx->range_text = range_text;
    ctx->range_dup = range_dup;
    ctx->doc_all_range = doc_all_range;

    ctx->selected_range = word_selected_range;
    ctx->range_find_proc = word_range_find;
    ctx->range_move_point = word_range_move_point;
    ctx->range_select = word_range_select;

    ctx->get_doc = get_word_doc;

    return S_OK;
}



DECLARE_FUNCTIONS(word)
